# Copyright 2020 Dirk Thomas
# Licensed under the Apache License, Version 2.0

import argparse
import re


def argument_package_name(value):
    """
    Check if an argument is a valid package name.

    Used as a ``type`` callback in ``add_argument()`` calls.
    Package names starting with a dash must be prefixed with a space to avoid
    collisions with command line arguments.

    :param str value: The command line argument
    :returns: The package name
    :raises argparse.ArgumentTypeError: if the value starts with a dash
    """
    if value.startswith('-'):
        raise argparse.ArgumentTypeError('unrecognized argument: ' + value)
    return value.lstrip()


def argument_valid_regex(value):
    """
    Check if an argument is a valid regular expression.

    Used as a ``type`` callback in ``add_argument()`` calls.

    :param str value: The command line argument
    :returns: The regular expression object
    :raises argparse.ArgumentTypeError: if the value is not a valid regex
    """
    try:
        return re.compile(value)
    except re.error as e:  # noqa: F841
        raise argparse.ArgumentTypeError(
            'must be a valid regex: {e}'.format_map(locals()))



# Copyright 2019 Dirk Thomas
# Licensed under the Apache License, Version 2.0

from colcon_core.event.job import JobEnded
from colcon_core.event.test import TestFailure
from colcon_core.event_handler import EventHandlerExtensionPoint
from colcon_core.plugin_system import satisfies_version
from colcon_core.verb.build import BuildPackageArguments
from colcon_core.verb.test import TestPackageArguments
from colcon_package_selection.package_selection.previous \
    import set_result
from colcon_package_selection.package_selection.previous \
    import TEST_FAILURE_RESULT


class StoreResultEventHandler(EventHandlerExtensionPoint):
    """
    Persist the result of a job in a file in its build directory.

    The extension handles events of the following types:
    - :py:class:`colcon_core.event.job.JobEnded`
    - :py:class:`colcon_core.event.test.TestFailure`
    """

    def __init__(self):  # noqa: D107
        super().__init__()
        satisfies_version(
            EventHandlerExtensionPoint.EXTENSION_POINT_VERSION, '^1.0')
        self._test_failures = set()

    def __call__(self, event):  # noqa: D102
        data = event[0]

        if isinstance(data, TestFailure):
            job = event[1]
            self._test_failures.add(job)

        elif isinstance(data, JobEnded):
            job = event[1]

            if isinstance(job.task_context.args, BuildPackageArguments):
                verb_name = 'build'
            elif isinstance(job.task_context.args, TestPackageArguments):
                verb_name = 'test'
            else:
                return

            if job in self._test_failures:
                result = TEST_FAILURE_RESULT
            else:
                result = data.rc

            set_result(job.task_context.args.build_base, verb_name, result)



# Copyright 2018 Dirk Thomas
# Licensed under the Apache License, Version 2.0

import re
import typing

from colcon_core.package_augmentation import PackageAugmentationExtensionPoint
from colcon_core.package_discovery import logger
from colcon_core.package_discovery import PackageDiscoveryExtensionPoint
from colcon_core.plugin_system import satisfies_version
from colcon_package_selection.argument import argument_package_name
from colcon_package_selection.argument import argument_valid_regex


class IgnorePackageDiscovery(
    PackageDiscoveryExtensionPoint, PackageAugmentationExtensionPoint,
):
    """Ignore discovered packages based on cli arguments."""

    def __init__(self):  # noqa: D107
        super().__init__()
        satisfies_version(
            PackageDiscoveryExtensionPoint.EXTENSION_POINT_VERSION, '^1.0')
        satisfies_version(
            PackageAugmentationExtensionPoint.EXTENSION_POINT_VERSION, '^1.0')
        self._args = None

    def add_arguments(self, *, parser, with_default):  # noqa: D102
        parser.add_argument(
            '--packages-ignore', nargs='*', metavar='PKG_NAME',
            type=argument_package_name,
            help='Ignore packages as if they were not discovered')
        parser.add_argument(
            '--packages-ignore-regex', nargs='*', metavar='PATTERN',
            type=argument_valid_regex,
            help='Ignore packages where any of the patterns match the package '
                 'name')

    def has_parameters(self, *, args):  # noqa: D102
        self._args = args
        return False

    def discover(self, *, args, identification_extensions):  # noqa: D102
        return set()

    def augment_packages(
        self, descs, *, additional_argument_names=None
    ):  # noqa: D102
        pkg_names = {d.name for d in descs}

        # check patterns and remove invalid ones
        for pattern in (self._args.packages_ignore_regex or []):
            if not any(re.match(pattern, pkg_name) for pkg_name in pkg_names):
                if isinstance(pattern, typing.Pattern):
                    pattern = pattern.pattern
                logger.warning(
                    "the --packages-ignore-regex '{pattern}' doesn't match "
                    'any of the package names'.format_map(locals()))

        for pkg_name in (self._args.packages_ignore or []):
            if pkg_name not in pkg_names:
                logger.warning(
                    "ignoring unknown package '{pkg_name}' in "
                    '--packages-ignore'.format_map(locals()))

        # remove the descriptors which should be ignored
        for desc in set(descs):
            if desc.name in (self._args.packages_ignore or []):
                descs.remove(desc)
                continue
            for pattern in (self._args.packages_ignore_regex or []):
                if re.match(pattern, desc.name):
                    descs.remove(desc)
                    break



# Copyright 2016-2018 Dirk Thomas
# Licensed under the Apache License, Version 2.0

import argparse
import re
import sys
import typing

from colcon_core.package_selection import logger
from colcon_core.package_selection import PackageSelectionExtensionPoint
from colcon_core.plugin_system import satisfies_version
from colcon_package_selection.argument import argument_package_name
from colcon_package_selection.argument import argument_valid_regex


class _DepthAndPackageNames(argparse.Action):
    """Action to assign an integer depth optional package names."""

    def __call__(self, parser, namespace, values, option_string=None):
        assert len(values) >= 1
        try:
            values[0] = int(values[0])
            if values[0] < 0:
                raise ValueError()
        except ValueError:
            raise argparse.ArgumentError(
                self, 'the first parameter must be a non-negative integer for '
                'the depth')
        for i in range(1, len(values)):
            values[i] = argument_package_name(values[i])
        setattr(namespace, self.dest, values)


class DependenciesPackageSelection(PackageSelectionExtensionPoint):
    """Select packages based on their dependencies."""

    def __init__(self):  # noqa: D107
        super().__init__()
        satisfies_version(
            PackageSelectionExtensionPoint.EXTENSION_POINT_VERSION, '^1.0')

    def add_arguments(self, *, parser):  # noqa: D102
        parser.add_argument(
            '--packages-up-to', nargs='*', metavar='PKG_NAME',
            type=argument_package_name,
            help='Only process a subset of packages and their recursive '
                 'dependencies')
        parser.add_argument(
            '--packages-up-to-regex', nargs='*', metavar='PATTERN',
            type=argument_valid_regex,
            help='Only process a subset of packages and their '
                 'recursive dependencies, where any of the '
                 'patterns match the package name')
        parser.add_argument(
            '--packages-above', nargs='*', metavar='PKG_NAME',
            type=argument_package_name,
            help='Only process a subset of packages and packages which '
                 'recursively depend on them')
        parser.add_argument(
            '--packages-above-and-dependencies', nargs='*', metavar='PKG_NAME',
            type=argument_package_name,
            help='Only process a subset of packages and packages which '
                 'recursively depend on them including all their recursive '
                 'dependencies')
        parser.add_argument(
            '--packages-above-depth', nargs='+',
            metavar=('DEPTH', 'PKG_NAME'), action=_DepthAndPackageNames,
            help='Only process a subset of packages and packages which '
                 'recursively depend on them up to a given depth')

        parser.add_argument(
            '--packages-select-by-dep', nargs='*', metavar='PKG_NAME',
            type=argument_package_name,
            help='Only process packages which (recursively) depend on this')
        parser.add_argument(
            '--packages-skip-by-dep', nargs='*', metavar='PKG_NAME',
            type=argument_package_name,
            help='Skip packages which (recursively) depend on this')
        parser.add_argument(
            '--packages-skip-up-to', nargs='*', metavar='PKG_NAME',
            type=argument_package_name,
            help='Skip a subset of packages and their recursive dependencies')

    def check_parameters(self, args, pkg_names):  # noqa: D102
        # exit on invalid arguments
        error_messages = []
        for name in args.packages_up_to or set():
            if name not in pkg_names:
                error_messages.append(
                    "Package '{name}' specified with --packages-up-to "
                    'was not found'
                    .format_map(locals()))
        for pattern in (args.packages_up_to_regex or []):
            if not any(re.match(pattern, pkg_name) for pkg_name in pkg_names):
                if isinstance(pattern, typing.Pattern):
                    pattern = pattern.pattern
                error_messages.append(
                    "the --packages-up-to-regex '{pattern}' doesn't match "
                    'any of the package names'.format_map(locals()))
        for name in args.packages_above or set():
            if name not in pkg_names:
                error_messages.append(
                    "Package '{name}' specified with --packages-above "
                    'was not found'
                    .format_map(locals()))
        for name in args.packages_above_and_dependencies or set():
            if name not in pkg_names:
                error_messages.append(
                    "Package '{name}' specified with "
                    '--packages-above-and-dependencies was not found'
                    .format_map(locals()))
        for name in (args.packages_above_depth or [])[1:]:
            if name not in pkg_names:
                error_messages.append(
                    "Package '{name}' specified with "
                    '--packages-above-depth was not found'
                    .format_map(locals()))
        if error_messages:
            sys.exit('\n'.join(error_messages))

    def select_packages(self, args, decorators):  # noqa: D102
        if args.packages_up_to:
            select_pkgs = set(args.packages_up_to)
            for decorator in reversed(decorators):
                if decorator.descriptor.name in select_pkgs:
                    select_pkgs |= set(decorator.recursive_dependencies)
                elif decorator.selected:
                    pkg = decorator.descriptor
                    logger.info(
                        "Skipping package '{pkg.name}' in '{pkg.path}'"
                        .format_map(locals()))
                    decorator.selected = False

        if args.packages_up_to_regex:
            select_pkgs = set()
            patterns = args.packages_up_to_regex
            for decorator in reversed(decorators):
                pkg = decorator.descriptor
                if (
                    pkg.name in select_pkgs or
                    any(re.match(pattern, pkg.name) for pattern in patterns)
                ):
                    select_pkgs |= set(decorator.recursive_dependencies)
                elif decorator.selected:
                    logger.info(
                        "Skipping package '{pkg.name}' in '{pkg.path}'"
                        .format_map(locals()))
                    decorator.selected = False

        if args.packages_above:
            select_pkgs = set(args.packages_above)
            for decorator in decorators:
                if decorator.descriptor.name in select_pkgs:
                    continue
                if not (select_pkgs & set(decorator.recursive_dependencies)):
                    if decorator.selected:
                        pkg = decorator.descriptor
                        logger.info(
                            "Skipping package '{pkg.name}' in '{pkg.path}'"
                            .format_map(locals()))
                        decorator.selected = False

        if args.packages_above_and_dependencies:
            # collect all above packages
            select_pkgs = set(args.packages_above_and_dependencies)
            for decorator in decorators:
                if decorator.descriptor.name in select_pkgs:
                    continue
                if set(decorator.recursive_dependencies) & select_pkgs:
                    select_pkgs.add(decorator.descriptor.name)

            for decorator in reversed(decorators):
                if decorator.descriptor.name in select_pkgs:
                    select_pkgs |= set(decorator.recursive_dependencies)
                elif decorator.selected:
                    pkg = decorator.descriptor
                    logger.info(
                        "Skipping package '{pkg.name}' in '{pkg.path}'"
                        .format_map(locals()))
                    decorator.selected = False

        if args.packages_above_depth and len(args.packages_above_depth) > 1:
            depth = args.packages_above_depth[0]
            select_pkgs = set(args.packages_above_depth[1:])
            for decorator in decorators:
                if decorator.descriptor.name in select_pkgs:
                    continue
                if not [
                    d for d in set(decorator.recursive_dependencies)
                    if d in select_pkgs and d.metadata['depth'] <= depth
                ]:
                    if decorator.selected:
                        pkg = decorator.descriptor
                        logger.info(
                            "Skipping package '{pkg.name}' in '{pkg.path}'"
                            .format_map(locals()))
                        decorator.selected = False

        if args.packages_select_by_dep:
            deps = set(args.packages_select_by_dep)
            for decorator in decorators:
                if not (deps & set(decorator.recursive_dependencies)):
                    if decorator.selected:
                        pkg = decorator.descriptor
                        logger.info(
                            "Skipping package '{pkg.name}' in '{pkg.path}'"
                            .format_map(locals()))
                        decorator.selected = False

        if args.packages_skip_by_dep:
            deps = set(args.packages_skip_by_dep)
            for decorator in decorators:
                if deps & set(decorator.recursive_dependencies):
                    if decorator.selected:
                        pkg = decorator.descriptor
                        logger.info(
                            "Skipping package '{pkg.name}' in '{pkg.path}'"
                            .format_map(locals()))
                        decorator.selected = False

        if args.packages_skip_up_to:
            skip_pkgs = set(args.packages_skip_up_to)
            for decorator in reversed(decorators):
                if decorator.descriptor.name in skip_pkgs:
                    skip_pkgs |= set(decorator.recursive_dependencies)
                    if decorator.selected:
                        pkg = decorator.descriptor
                        logger.info(
                            "Skipping package '{pkg.name}' in '{pkg.path}'"
                            .format_map(locals()))
                        decorator.selected = False



# Copyright 2016-2018 Dirk Thomas
# Licensed under the Apache License, Version 2.0

import sys

from colcon_core.package_selection import logger
from colcon_core.package_selection import PackageSelectionExtensionPoint
from colcon_core.plugin_system import satisfies_version
from colcon_package_selection.argument import argument_package_name


class StartEndPackageSelection(PackageSelectionExtensionPoint):
    """Select a range of packages based on flattened topological ordering."""

    def __init__(self):  # noqa: D107
        super().__init__()
        satisfies_version(
            PackageSelectionExtensionPoint.EXTENSION_POINT_VERSION, '^1.0')

    def add_arguments(self, *, parser):  # noqa: D102
        parser.add_argument(
            '--packages-start', metavar='PKG_NAME',
            type=argument_package_name,
            help='Skip packages before this in flat topological ordering')
        parser.add_argument(
            '--packages-end', metavar='PKG_NAME',
            type=argument_package_name,
            help='Skip packages after this in flat topological ordering')

    def check_parameters(self, args, pkg_names):  # noqa: D102
        # exit on invalid arguments
        if args.packages_start and args.packages_start not in pkg_names:
            sys.exit(
                "Package '{args.packages_start}' specified with "
                '--packages-start was not found'
                .format_map(locals()))
        if args.packages_end and args.packages_end not in pkg_names:
            sys.exit(
                "Package '{args.packages_end}' specified with --packages-end "
                'was not found'
                .format_map(locals()))

    def select_packages(self, args, decorators):  # noqa: D102
        pkg_within_range = not args.packages_start
        for decorator in decorators:
            pkg = decorator.descriptor

            # identify start of range
            if pkg.name == args.packages_start:
                pkg_within_range = True

            selected = pkg_within_range
            if decorator.selected and not selected:
                # mark packages outside of the range as not selected
                logger.info(
                    "Skipping package '{pkg.name}' in '{pkg.path}'"
                    .format_map(locals()))
                decorator.selected = False

            # identify end of range
            if pkg.name == args.packages_end:
                if not pkg_within_range:
                    sys.exit(
                        "The --packages-end package '{args.packages_end}' "
                        'occurs topologically before the --packages-start '
                        "package '{args.packages_start}'"
                        .format_map(locals()))
                pkg_within_range = False



# Copyright 2019 Dirk Thomas
# Licensed under the Apache License, Version 2.0

import os

from colcon_core.package_selection import logger
from colcon_core.package_selection import PackageSelectionExtensionPoint
from colcon_core.plugin_system import satisfies_version
from colcon_core.subprocess import SIGINT_RESULT
from colcon_package_selection.package_selection.previous \
    import get_previous_result


class PreviousPackageSelectionExtension(PackageSelectionExtensionPoint):
    """Skip a set of packages based on results of previous invocations."""

    def __init__(self):  # noqa: D107
        super().__init__()
        satisfies_version(
            PackageSelectionExtensionPoint.EXTENSION_POINT_VERSION, '^1.0')

    def add_arguments(self, *, parser):  # noqa: D102
        group = parser.add_mutually_exclusive_group()
        group.add_argument(
            '--packages-select-build-failed', action='store_true',
            help='Only process a subset of packages which have failed to '
                 'build previously (aborted packages are not '
                 'considered errors)')
        group.add_argument(
            '--packages-skip-build-finished', action='store_true',
            help='Skip a set of packages which have finished to build '
                 'previously')
        group.add_argument(
            '--packages-select-test-failures', action='store_true',
            help='Only process a subset of packages which had test failures '
                 'previously')
        group.add_argument(
            '--packages-skip-test-passed', action='store_true',
            help='Skip a set of packages which had no test failures '
                 'previously')

    def select_packages(self, args, decorators):  # noqa: D102
        if not any((
            args.packages_select_build_failed,
            args.packages_skip_build_finished,
            args.packages_select_test_failures,
            args.packages_skip_test_passed,
        )):
            return

        if not hasattr(args, 'build_base'):
            if args.packages_select_build_failed:
                argument = '--packages-select-build-failed'
            elif args.packages_skip_build_finished:
                argument = '--packages-skip-build-finished'
            elif args.packages_select_test_failures:
                argument = '--packages-select-test-failures'
            elif args.packages_skip_test_passed:
                argument = '--packages-skip-test-passed'
            else:
                assert False
            logger.warning(
                "Ignoring '{argument}' since the invoked verb doesn't have a "
                "'--build-base' argument and therefore can't access "
                'information about the previous state of a package'
                .format_map(locals()))
            return

        for decorator in decorators:
            # skip packages which have already been ruled out
            if not decorator.selected:
                continue

            pkg = decorator.descriptor

            if (
                args.packages_select_build_failed or
                args.packages_skip_build_finished
            ):
                verb_name = 'build'
            elif (
                args.packages_select_test_failures or
                args.packages_skip_test_passed
            ):
                verb_name = 'test'
            else:
                assert False

            previous_result = get_previous_result(
                os.path.join(args.build_base, pkg.name), verb_name)

            if args.packages_select_build_failed:
                package_kind = None
                if previous_result is None:
                    package_kind = 'not previously built'
                elif previous_result == SIGINT_RESULT:
                    package_kind = 'previously aborted'
                elif previous_result == '0':
                    package_kind = 'previously built'
                if package_kind is not None:
                    logger.info(
                        "Skipping {package_kind} package '{pkg.name}' in "
                        "'{pkg.path}'".format_map(locals()))
                    decorator.selected = False

            if args.packages_skip_build_finished:
                if previous_result == '0':
                    logger.info(
                        "Skipping previously built package '{pkg.name}' in "
                        "'{pkg.path}'".format_map(locals()))
                    decorator.selected = False

            if args.packages_select_test_failures:
                package_kind = None
                if previous_result is None:
                    package_kind = 'not previously tested'
                elif previous_result == SIGINT_RESULT:
                    package_kind = 'previously aborted'
                elif previous_result == '0':
                    package_kind = 'previously tested'
                if package_kind is not None:
                    logger.info(
                        "Skipping {package_kind} package '{pkg.name}' in "
                        "'{pkg.path}'".format_map(locals()))
                    decorator.selected = False

            if args.packages_skip_test_passed:
                if previous_result == '0':
                    logger.info(
                        "Skipping previously tested package '{pkg.name}' in "
                        "'{pkg.path}'".format_map(locals()))
                    decorator.selected = False



# Copyright 2019 Dirk Thomas
# Licensed under the Apache License, Version 2.0

import pathlib

RESULT_FILENAME = 'colcon_{verb_name}.rc'

TEST_FAILURE_RESULT = 'test failures'


def get_previous_result(package_build_base, verb_name):
    """
    Get the result of a verb from the package build directory.

    :param str package_build_base: The build directory of a package
    :param str verb_name: The invoked verb name
    :returns: The previously persisted result, otherwise None
    :rtype: str
    """
    path = _get_result_path(package_build_base, verb_name)
    if not path.exists():
        return None
    return path.read_text().rstrip()


def set_result(package_build_base, verb_name, result):
    """
    Persist the result of a verb in the package build directory.

    :param str package_build_base: The build directory of a package
    :param str verb_name: The invoked verb name
    :param str result: The result of the invocation
    """
    path = _get_result_path(package_build_base, verb_name)
    path.parent.mkdir(parents=True, exist_ok=True)
    path.write_text(str(result) + '\n')


def _get_result_path(package_build_base, verb_name):
    return pathlib.Path(
        package_build_base) / RESULT_FILENAME.format_map(locals())



# Copyright 2016-2018 Dirk Thomas
# Licensed under the Apache License, Version 2.0

import re
import typing

from colcon_core.package_selection import logger
from colcon_core.package_selection import PackageSelectionExtensionPoint
from colcon_core.plugin_system import satisfies_version
from colcon_package_selection.argument import argument_package_name
from colcon_package_selection.argument import argument_valid_regex


class SelectSkipPackageSelectionExtension(PackageSelectionExtensionPoint):
    """Select a set of packages based on select / skip criteria."""

    def __init__(self):  # noqa: D107
        super().__init__()
        satisfies_version(
            PackageSelectionExtensionPoint.EXTENSION_POINT_VERSION, '^1.0')

    def add_arguments(self, *, parser):  # noqa: D102
        parser.add_argument(
            '--packages-select', nargs='*', metavar='PKG_NAME',
            type=argument_package_name,
            help='Only process a subset of packages')
        parser.add_argument(
            '--packages-skip', nargs='*', metavar='PKG_NAME',
            type=argument_package_name,
            help='Skip a set of packages')

        parser.add_argument(
            '--packages-select-regex', nargs='*', metavar='PATTERN',
            type=argument_valid_regex,
            help='Only process a subset of packages where any of the patterns '
                 'match the package name')
        parser.add_argument(
            '--packages-skip-regex', nargs='*', metavar='PATTERN',
            type=argument_valid_regex,
            help='Skip a set of packages where any of the patterns match the '
                 'package name')

    def check_parameters(self, args, pkg_names):  # noqa: D102
        # warn about ignored arguments
        for pkg_name in args.packages_select or []:
            if pkg_name not in pkg_names:
                logger.warning(
                    "ignoring unknown package '{pkg_name}' in "
                    '--packages-select'.format_map(locals()))

        for pkg_name in args.packages_skip or []:
            if pkg_name not in pkg_names:
                logger.warning(
                    "ignoring unknown package '{pkg_name}' in "
                    '--packages-skip'.format_map(locals()))

        for pattern in (args.packages_select_regex or []):
            if not any(re.match(pattern, pkg_name) for pkg_name in pkg_names):
                if isinstance(pattern, typing.Pattern):
                    pattern = pattern.pattern
                logger.warning(
                    "the --packages-select-regex '{pattern}' doesn't match "
                    'any of the package names'.format_map(locals()))

        for pattern in (args.packages_skip_regex or []):
            if not any(re.match(pattern, pkg_name) for pkg_name in pkg_names):
                if isinstance(pattern, typing.Pattern):
                    pattern = pattern.pattern
                logger.warning(
                    "the --packages-skip-regex '{pattern}' doesn't match any "
                    'of the package names'.format_map(locals()))

    def select_packages(self, args, decorators):  # noqa: D102
        for decorator in decorators:
            # skip packages which have already been ruled out
            if not decorator.selected:
                continue

            pkg = decorator.descriptor

            if (
                pkg.name in (args.packages_skip or []) or
                any(
                    re.match(pattern, pkg.name)
                    for pattern in (args.packages_skip_regex or [])
                )
            ):
                logger.info(
                    "Skipping package '{pkg.name}' in '{pkg.path}'"
                    .format_map(locals()))
                decorator.selected = False

            elif (
                args.packages_select is not None or
                args.packages_select_regex is not None
            ):
                if (
                    pkg.name not in (args.packages_select or []) and
                    not any(
                        re.match(pattern, pkg.name)
                        for pattern in (args.packages_select_regex or [])
                    )
                ):
                    logger.info(
                        "Skipping not selected package '{pkg.name}' in "
                        "'{pkg.path}'".format_map(locals()))
                    decorator.selected = False
